/* Copyright 2015 Jetperch LLC.  All rights reserved. */

/**
 * @file
 *
 * @brief Design by contract macros.
 */

#ifndef EMBC_DBC_H_
#define EMBC_DBC_H_


#include "log.h"
#include "ec.h"


/**
 * @ingroup embc
 * @defgroup dbc Design by contract
 *
 * @brief Macros and functions to support design by contract.
 *
 * Define dbc_assert() to provide appropriate assert handling for your
 * platform and application.
 *
 * References include:
 *
 * - http://dbc.rubyforge.org/
 * - https://www.google.com/url?sa=t&rct=j&q=&esrc=s&source=web&cd=3&cad=rja&uact=8&ved=0CCsQFjACahUKEwiLoK6g04HIAhWFmh4KHY8aAvo&url=http%3A%2F%2Fwww.barrgroup.com%2FEmbedded-Systems%2FHow-To%2FDesign-by-Contract-for-Embedded-Software&usg=AFQjCNHiKuGGWvikJrMCsyzwSx1y229nYQ&sig2=wFxNiKE3JQpYNfFZRif1tQ&bvm=bv.103073922,d.dmo
 *
 * @{
 */

/**
 * @def DBC_ARG_FAIL_RETURN_CODE
 *
 * @brief The return code for argument check failure.
 */
#ifndef DBC_ARG_FAIL_RETURN_CODE
#define DBC_ARG_FAIL_RETURN_CODE EMBC_ERROR_PARAMETER_INVALID
#endif

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

/**
 * @brief the function called on a design by contract assertion.
 *
 * @param file The source filename.
 * @param line The source filename's line number.
 * @param msg The message for the design-by-contract assertion.
 *
 * The library provides a weak definition of this function that does nothing.
 * Your application should appropriately define this function to signal
 * the assertion.  When embc logging is enabled, a log message is also
 * generated.
 */
void dbc_assert(char const *file, unsigned line, const char * msg);

#ifdef __cplusplus
}
#endif /* __cplusplus */

/**
 * @brief Assert on a design-by-contract condition.
 *
 * @param condition The condition for the assert.  The condition is successful
 *      when true.  When false, an error has occurred.
 * @param message The message to display when condition is false.
 */
#define DBC_ASSERT(condition, message) \
	if (!(condition)) { \
		LOGF_ERROR("dbc_assert: %s", message); \
        dbc_assert(__FILE__, __LINE__, message); \
    }

/**
 * @brief Assert on a null value.
 *
 * @param x The expression which should not be null.
 */
#define DBC_NOT_NULL(x) DBC_ASSERT((x) != 0, #x " is null")

/**
 * @defgroup arg Arguments
 *
 * @brief Argument checking macros which cause the enclosing function to
 *      return an error on assert failure.
 *
 * @{
 */


/**
 * @brief Assert on an invalid argument value and execute and action on error.
 *
 * @param condition The condition for the assert.  The condition is successful
 *      when true.  When false, an error has occurred.
 * @param action The action to execute when condition fails.
 * @param message The message to display when condition is false.
 *
 * SIDE EFFECT: This macro will cause the enclosing function to execute action
 *      on an assert.
 */
#define DBC_ASSERT_ACTION(condition, action, message) \
    if (!(condition)) { \
        LOGF_ERROR("dbc_assert: %s", message); \
        dbc_assert(__FILE__, __LINE__, message); \
        action; \
    }

/**
 * @brief Assert on an invalid argument value and return an error.
 *
 * @param condition The condition for the assert.  The condition is successful
 *      when true.  When false, an error has occurred.
 * @param message The message to display when condition is false.
 * @return Does not return on success, but this macro will cause the enclosing
 *      function to return ARG_FAIL_RETURN_CODE on an assert.
 */
#define DBC_ASSERT_RWE(condition, message) \
		DBC_ASSERT_ACTION(condition, return DBC_ARG_FAIL_RETURN_CODE, message);


/**
 * @brief Assert that a function argument is not null.
 *
 * @param x The function argument to check.
 */
#define DBC_ARG_NOT_NULL(x) DBC_ASSERT_RWE((x) != 0, #x " is null")

/**
 * @brief Assert that a function argument is greater than zero.
 *
 * @param x The function argument to check.
 */
#define DBC_ARG_GT_ZERO(x) DBC_ASSERT_RWE((x) > 0, #x " <= 0")

/**
 * @brief Assert that a function argument is greater than or equal to zero.
 *
 * @param x The function argument to check.
 */
#define DBC_ARG_GTE_ZERO(x) DBC_ASSERT_RWE((x) > 0, #x " < 0")

/**
 * @brief Assert that a function argument is not equal to zero.
 *
 * @param x The function argument to check.
 */
#define DBC_ARG_NE_ZERO(x) DBC_ASSERT_RWE((x) != 0, #x " != 0")

/**
 * @brief Assert that a function argument is less than zero.
 *
 * @param x The function argument to check.
 */
#define DBC_ARG_LT_ZERO(x) DBC_ASSERT_RWE((x) < 0, #x " >= 0")

/**
 * @brief Assert that a function argument is less than or equal to zero.
 *
 * @param x The function argument to check.
 */
#define DBC_ARG_LTE_ZERO(x) DBC_ASSERT_RWE((x) < 0, #x " > 0")

/**
 * @brief Assert on a function argument condition.
 *
 * @param x The function argument or condition to check.
 */
#define DBC_ARG_REQUIRE(x) DBC_ASSERT_RWE(x, #x)

/**
 * @brief Assert on a function argument condition.
 *
 * @param x The function argument or condition to check.
 */
#define DBC_ARG_ASSERT(x) DBC_ASSERT_RWE(x, #x)

/** @} */

/** @} */

#endif /* EMBC_DBC_H_ */
