// Copyright (c) 2014 Hillcrest Laboratories, Inc.  All rights reserved.

#include "CppUTest/TestHarness.h"
#include "embc/fifo.h"
#include "embc/ec.h"
#include <cstring>

#define ARRAY_SIZE(x) sizeof(x) / sizeof(x[0])
#define DATA_SIZE 1024

TEST_GROUP(FIFO8)
{
    struct fifo8_s * f;
    fifo_size_t sz;
    uint8_t data[DATA_SIZE];

    void setup() {
        f = 0;
        sz = 0;
        memset(data, 0, sizeof(data));
    }

    void teardown() {
        fifo8_free(f);
    }
};

TEST(FIFO8, alloc)
{
    CHECK_EQUAL(0, fifo8_alloc(&f, 16));
    CHECK_EQUAL(0, fifo8_size(f));
    CHECK_EQUAL(1, fifo8_is_empty(f));
    CHECK_EQUAL(0, fifo8_is_full(f));
}

TEST(FIFO8, one_byte) {
    uint8_t d1 = 0x42;
    uint8_t d2 = 0;
    fifo_size_t sz;
    CHECK_EQUAL(0, fifo8_alloc(&f, 16));
    CHECK_EQUAL(0, fifo8_write_u8(f, d1));
    CHECK_EQUAL(1, fifo8_size(f));
    CHECK_EQUAL(0, fifo8_is_empty(f));
    CHECK_EQUAL(0, fifo8_is_full(f));
    CHECK_EQUAL(0, fifo8_read_u8(f, &d2));
    BYTES_EQUAL(d1, d2);
    CHECK_EQUAL(0, fifo8_size(f));
}

TEST(FIFO8, read_empty) {
    uint8_t d2;
    CHECK_EQUAL(0, fifo8_alloc(&f, 16));
    CHECK_EQUAL(EMBC_ERROR_EMPTY, fifo8_read(f, &d2, 1, &sz));
    CHECK_EQUAL(0, sz);
    CHECK_EQUAL(EMBC_ERROR_EMPTY, fifo8_read_u8(f, &d2));
}

TEST(FIFO8, write_full) {
    uint8_t d2;
    CHECK_EQUAL(0, fifo8_alloc(&f, 16));
    for (int i = 0; i < 15; ++i) {
        CHECK_EQUAL(0, fifo8_write_u8(f, (uint8_t) i));
    }
    CHECK_EQUAL(0, fifo8_read_u8(f, &d2));
    CHECK_EQUAL(0, d2);
    CHECK_EQUAL(0, fifo8_is_full(f));
    CHECK_EQUAL(0, fifo8_write_u8(f, 15));
    CHECK_EQUAL(EMBC_ERROR_FULL, fifo8_write_u8(f, 0x43));
    CHECK_EQUAL(1, fifo8_is_full(f));
    for (int i = 1; i < 16; ++i) {
        CHECK_EQUAL(0, fifo8_read_u8(f, &d2));
        CHECK_EQUAL(i, d2);
    }
    CHECK_EQUAL(EMBC_ERROR_EMPTY, fifo8_read_u8(f, &d2));
}

TEST(FIFO8, bytes) {
    uint8_t d1[] = {0x42, 0x43, 0x44};
    uint8_t d2[sizeof(d1)];
    fifo_size_t sz;
    CHECK_EQUAL(0, fifo8_alloc(&f, 16));
    CHECK_EQUAL(0, fifo8_write(f, d1, sizeof(d1), &sz));
    CHECK_EQUAL(3, sz);
    CHECK_EQUAL(3, fifo8_size(f));

    CHECK_EQUAL(0, fifo8_read(f, d2, sizeof(d2), &sz));
    CHECK_EQUAL(sizeof(d2), sz);
    BYTES_EQUAL(d1[0], d2[0]);
    BYTES_EQUAL(d1[1], d2[1]);
    BYTES_EQUAL(d1[2], d2[2]);
}

TEST(FIFO8, wrap) {
    uint8_t d;
    CHECK_EQUAL(0, fifo8_alloc(&f, 16));
    for (int i = 0; i < 32; ++i) {
        CHECK_EQUAL(0, fifo8_write_u8(f, (uint8_t) i));
        CHECK_EQUAL(1, fifo8_size(f));
        CHECK_EQUAL(0, fifo8_read_u8(f, &d));
        CHECK_EQUAL(i, d);
        CHECK_EQUAL(0, fifo8_size(f));
    }
}
